#include <algorithm>
#include <iostream>
#include <fstream>
#include <limits>
#include <stdint.h>
#include <boost/lexical_cast.hpp>
#include <boost/filesystem.hpp>

#include "parallel/MeshDistributor.h"
#include "parallel/MeshManipulation.h"
#include "parallel/ParallelDebug.h"
#include "parallel/StdMpi.h"
#include "parallel/ParMetisPartitioner.h"
#include "io/ElementFileWriter.h"
#include "io/MacroInfo.h"
#include "io/VtkWriter.h"
#include "Mesh.h"
#include "Traverse.h"
#include "ElInfo.h"
#include "Element.h"
#include "MacroElement.h"
#include "DOFMatrix.h"
#include "DOFVector.h"
#include "SystemVector.h"
#include "ElementDofIterator.h"
#include "ProblemStatBase.h"
#include "StandardProblemIteration.h"
#include "VertexVector.h"
#include "MeshStructure.h"
#include "ProblemVec.h"
#include "ProblemInstat.h"
#include "Debug.h"

namespace AMDiS {

  using boost::lexical_cast;
  using namespace boost::filesystem;

  inline bool cmpDofsByValue(const DegreeOfFreedom* dof1, const DegreeOfFreedom* dof2)
  {
    return (*dof1 < *dof2);
  }


  MeshDistributor::MeshDistributor(std::string str)
    : probStat(0),
      name(str),
      feSpace(NULL),
      mesh(NULL),
      refineManager(NULL),
      info(10),
      partitioner(NULL),
      nRankDofs(0),
      nOverallDofs(0),
      rstart(0),
      deserialized(false),
      writeSerializationFile(false),
      repartitioningAllowed(false),
      nTimestepsAfterLastRepartitioning(0),
      repartCounter(0),
      lastMeshChangeIndex(0)
  {
    FUNCNAME("MeshDistributor::ParalleDomainBase()");

    mpiRank = MPI::COMM_WORLD.Get_rank();
    mpiSize = MPI::COMM_WORLD.Get_size();
    mpiComm = MPI::COMM_WORLD;

    int tmp = 0;
    GET_PARAMETER(0, name + "->repartitioning", "%d", &tmp);
    repartitioningAllowed = (tmp > 0);
  }


  void MeshDistributor::initParallelization()
  {
    FUNCNAME("MeshDistributor::initParallelization()");

    TEST_EXIT(mpiSize > 1)
      ("Parallelization does not work with only one process!\n");

    TEST_EXIT(feSpace)("No FE space has been defined for the mesh distributor!\n");
    TEST_EXIT(mesh)("No mesh has been defined for the mesh distributor!\n");

    // If the problem has been already read from a file, we need only to set
    // isRankDofs to all matrices and rhs vector and to remove periodic 
    // boundary conditions (if there are some).
    if (deserialized) {
      setRankDofs();
      removePeriodicBoundaryConditions();

      macroElIndexMap.clear();
      macroElIndexTypeMap.clear();
      std::deque<MacroElement*>& macros = mesh->getMacroElements();
      for (std::deque<MacroElement*>::iterator it = macros.begin();
	   it != macros.end(); ++it) {
	macroElIndexMap[(*it)->getIndex()] = (*it)->getElement();
	macroElIndexTypeMap[(*it)->getIndex()] = (*it)->getElType();
      }

      return;
    }

   
    // Test, if the mesh is the macro mesh only! Paritioning of the mesh is supported
    // only for macro meshes, so it will not work yet if the mesh is already refined
    // in some way.
    testForMacroMesh();

    // For later mesh repartitioning, we need to store some information about the
    // macro mesh.
    createMacroElementInfo();

    // create an initial partitioning of the mesh
    partitioner->createPartitionData();

    // set the element weights, which are 1 at the very first begin
    setInitialElementWeights();

    // and now partition the mesh    
    partitioner->fillCoarsePartitionVec(&oldPartitionVec);
    partitioner->partition(&elemWeights, INITIAL);
    partitioner->fillCoarsePartitionVec(&partitionVec);


#if (DEBUG != 0)
    debug::ElementIdxToDofs elMap;
    debug::createSortedDofs(mesh, elMap);
    if (mpiRank == 0) {
      int writePartMesh = 1;
      GET_PARAMETER(0, "dbg->write part mesh", "%d", &writePartMesh);

      if (writePartMesh > 0) {
	debug::writeElementIndexMesh(mesh, "elementIndex.vtu");
	writePartitioningMesh("part.vtu");
      } else {
	MSG("Skip write part mesh!\n");
      }
    }
#endif


    // === Create interior boundary information. ===

    createInteriorBoundaryInfo();

#if (DEBUG != 0)
    ParallelDebug::printBoundaryInfo(*this);
#endif


    // === Remove all macro elements that are not part of the rank partition. ===

    removeMacroElements();


    // === Create new global and local DOF numbering. ===


    // We have to remove the VertexVectors, which contain periodic assoiciations, 
    // because they are not valid anymore after some macro elements have been removed
    // and the corresponding DOFs were deleted.
    for (std::map<BoundaryType, VertexVector*>::iterator it = mesh->getPeriodicAssociations().begin();
	 it != mesh->getPeriodicAssociations().end(); ++it)
      const_cast<DOFAdmin&>(mesh->getDofAdmin(0)).removeDOFContainer(dynamic_cast<DOFContainer*>(it->second));

    updateLocalGlobalNumbering();


    // === If in debug mode, make some tests. ===

#if (DEBUG != 0)
    MSG("AMDiS runs in debug mode, so make some test ...\n");

    ParallelDebug::testAllElements(*this);
    debug::testSortedDofs(mesh, elMap);
    ParallelDebug::testInteriorBoundary(*this);
    ParallelDebug::testCommonDofs(*this, true);
    ParallelDebug::testGlobalIndexByCoords(*this);

    debug::writeMesh(feSpace, -1, "macro_mesh");   

    MSG("Debug mode tests finished!\n");
#endif


    // === Create periodic dof mapping, if there are periodic boundaries. ===

    createPeriodicMap();    

    // === Global refinements. ===

    int globalRefinement = 0;
    GET_PARAMETER(0, mesh->getName() + "->global refinements", "%d", &globalRefinement);

    if (globalRefinement > 0) {
      refineManager->globalRefine(mesh, globalRefinement);

#if (DEBUG != 0)
      debug::writeMesh(feSpace, -1, "gr_mesh");
#endif

      updateLocalGlobalNumbering();

     
      // === Update periodic mapping, if there are periodic boundaries. ===
      
      createPeriodicMap();
    }


    /// === Set DOF rank information to all matrices and vectors. ===

    setRankDofs();


    // === Remove periodic boundary conditions in sequential problem definition. ===

    removePeriodicBoundaryConditions();
  }


  void MeshDistributor::addProblemStat(ProblemVec *probVec)
  {
    FUNCNAME("MeshDistributor::addProblemVec()");

    if (feSpace != NULL) {
      std::vector<FiniteElemSpace*> feSpaces = probVec->getFeSpaces();
      for (unsigned int i = 0; i < feSpaces.size(); i++) {
	TEST_EXIT(feSpace == feSpaces[i])
	  ("Parallelizaton is not supported for multiple FE spaces!\n");
      }
    } else {
      feSpace = probVec->getFeSpace(0);
      mesh = feSpace->getMesh();
      info = probVec->getInfo();
      
      TEST_EXIT(mesh->getNumberOfDOFAdmin() == 1)
	("Only meshes with one DOFAdmin are supported!\n");
      TEST_EXIT(mesh->getDofAdmin(0).getNumberOfPreDOFs(0) == 0)
	("Wrong pre dof number for DOFAdmin!\n");
      
      switch (mesh->getDim()) {
      case 2:
	refineManager = new RefinementManager2d();
	break;
      case 3:
	refineManager = new RefinementManager3d();
	break;
      default:
	ERROR_EXIT("This should not happen for dim = %d!\n", mesh->getDim());
      }

      partitioner = new ParMetisPartitioner(mesh, &mpiComm);
    }

    // Create parallel serialization file writer, if needed.
    int writeSerialization = 0;
    GET_PARAMETER(0, probVec->getName() + "->output->write serialization", "%d", 
		  &writeSerialization);
    if (writeSerialization && !writeSerializationFile) {
      std::string filename = "";
      GET_PARAMETER(0, name + "->output->serialization filename", &filename);
      
      TEST_EXIT(filename != "")
	("No filename defined for parallel serialization file!\n");

      int tsModulo = -1;
      GET_PARAMETER(0, probVec->getName() + "->output->write every i-th timestep", 
		    "%d", &tsModulo);
      
      probVec->getFileWriterList().push_back(new Serializer<MeshDistributor>(this, filename, tsModulo));
      writeSerializationFile = true;
    }    

    int readSerialization = 0;
    GET_PARAMETER(0, probVec->getName() + "->input->read serialization", "%d", 
		  &readSerialization);
    if (readSerialization) {
      std::string filename = "";
      GET_PARAMETER(0, probVec->getName() + "->input->serialization filename", &filename);
      filename += ".p" + lexical_cast<std::string>(mpiRank);
      MSG("Start deserialization with %s\n", filename.c_str());
      std::ifstream in(filename.c_str());

      TEST_EXIT(!in.fail())("Could not open deserialization file: %s\n",
			    filename.c_str());

      probVec->deserialize(in);
      in.close();
      MSG("Deserialization from file: %s\n", filename.c_str());

      filename = "";
      GET_PARAMETER(0, name + "->input->serialization filename", &filename);
      
      TEST_EXIT(filename != "")
	("No filename defined for parallel deserialization file!\n");
      
      std::string rankFilename = filename + ".p" + lexical_cast<std::string>(mpiRank);
      in.open(rankFilename.c_str());
      
      TEST_EXIT(!in.fail())("Could not open parallel deserialization file: %s\n",
			    filename.c_str());
      
      deserialize(in);
      in.close();
      MSG("Deserializtion of mesh distributor from file: %s\n", rankFilename.c_str());
      deserialized = true;
    }

    probStat.push_back(probVec);
  }


  void MeshDistributor::exitParallelization()
  {}

  
  void MeshDistributor::testForMacroMesh()
  {
    FUNCNAME("MeshDistributor::testForMacroMesh()");

    int nMacroElements = 0;

    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, -1, Mesh::CALL_LEAF_EL);
    while (elInfo) {
      TEST_EXIT(elInfo->getLevel() == 0)
	("Mesh is already refined! This does not work with parallelization!\n");

      TEST_EXIT(elInfo->getType() == 0)
	("Only macro elements with level 0 are supported!\n");
      
      nMacroElements++;

      elInfo = stack.traverseNext(elInfo);
    }

    TEST_EXIT(nMacroElements >= mpiSize)
      ("The mesh has less macro elements than number of mpi processes!\n");
  }


  void MeshDistributor::synchVector(DOFVector<double> &vec)
  {
    StdMpi<std::vector<double> > stdMpi(mpiComm);

    for (RankToDofContainer::iterator sendIt = sendDofs.begin();
	 sendIt != sendDofs.end(); ++sendIt) {
      std::vector<double> dofs;
      int nSendDofs = sendIt->second.size();
      dofs.reserve(nSendDofs);
      
      for (int i = 0; i < nSendDofs; i++)
	dofs.push_back(vec[*((sendIt->second)[i])]);

      stdMpi.send(sendIt->first, dofs);
    }

    for (RankToDofContainer::iterator recvIt = recvDofs.begin();
	 recvIt != recvDofs.end(); ++recvIt)
      stdMpi.recv(recvIt->first, recvIt->second.size());

    stdMpi.startCommunication<double>(MPI_DOUBLE);

    for (RankToDofContainer::iterator recvIt = recvDofs.begin();
	 recvIt != recvDofs.end(); ++recvIt)
      for (unsigned int i = 0; i < recvIt->second.size(); i++)
	vec[*(recvIt->second)[i]] = stdMpi.getRecvData(recvIt->first)[i];
  }


  void MeshDistributor::synchVector(SystemVector &vec)
  {
    int nComponents = vec.getSize();
    StdMpi<std::vector<double> > stdMpi(mpiComm);

    for (RankToDofContainer::iterator sendIt = sendDofs.begin();
	 sendIt != sendDofs.end(); ++sendIt) {
      std::vector<double> dofs;
      int nSendDofs = sendIt->second.size();
      dofs.reserve(nComponents * nSendDofs);
      
      for (int i = 0; i < nComponents; i++) {
	DOFVector<double> *dofvec = vec.getDOFVector(i);
	for (int j = 0; j < nSendDofs; j++)
	  dofs.push_back((*dofvec)[*((sendIt->second)[j])]);
      }

      stdMpi.send(sendIt->first, dofs);
    }

    for (RankToDofContainer::iterator recvIt = recvDofs.begin();
	 recvIt != recvDofs.end(); ++recvIt)
      stdMpi.recv(recvIt->first, recvIt->second.size() * nComponents);

    stdMpi.startCommunication<double>(MPI_DOUBLE);

    for (RankToDofContainer::iterator recvIt = recvDofs.begin();
	 recvIt != recvDofs.end(); ++recvIt) {
      int nRecvDofs = recvIt->second.size();

      int counter = 0;
      for (int i = 0; i < nComponents; i++) {
	DOFVector<double> *dofvec = vec.getDOFVector(i);
 	for (int j = 0; j < nRecvDofs; j++)
	  (*dofvec)[*(recvIt->second)[j]] = 
	    stdMpi.getRecvData(recvIt->first)[counter++];
      }
    }
  }


  void MeshDistributor::setRankDofs()
  {
    for (unsigned int i = 0; i < probStat.size(); i++) {
      int nComponents = probStat[i]->getNumComponents();
      for (int j = 0; j < nComponents; j++) {
	for (int k = 0; k < nComponents; k++)
	  if (probStat[i]->getSystemMatrix(j, k))
	    probStat[i]->getSystemMatrix(j, k)->setRankDofs(isRankDof);

	TEST_EXIT_DBG(probStat[i]->getRhs()->getDOFVector(j))("No RHS vector!\n");
	TEST_EXIT_DBG(probStat[i]->getSolution()->getDOFVector(j))("No solution vector!\n");
	
	probStat[i]->getRhs()->getDOFVector(j)->setRankDofs(isRankDof);
	probStat[i]->getSolution()->getDOFVector(j)->setRankDofs(isRankDof);
      }
    }
  }


  void MeshDistributor::removePeriodicBoundaryConditions()
  {
    FUNCNAME("MeshDistributor::removePeriodicBoundaryConditions()");

    // Remove periodic boundaries in boundary manager on matrices and vectors.
    for (unsigned int i = 0; i < probStat.size(); i++) {
      int nComponents = probStat[i]->getNumComponents();

      for (int j = 0; j < nComponents; j++) {
	for (int k = 0; k < nComponents; k++) {
	  DOFMatrix* mat = probStat[i]->getSystemMatrix(j, k);
	  if (mat && mat->getBoundaryManager())
	    removePeriodicBoundaryConditions(const_cast<BoundaryIndexMap&>(mat->getBoundaryManager()->getBoundaryConditionMap()));
	}
	
	if (probStat[i]->getSolution()->getDOFVector(j)->getBoundaryManager())
	  removePeriodicBoundaryConditions(const_cast<BoundaryIndexMap&>(probStat[i]->getSolution()->getDOFVector(j)->getBoundaryManager()->getBoundaryConditionMap()));
	
	if (probStat[i]->getRhs()->getDOFVector(j)->getBoundaryManager())
	  removePeriodicBoundaryConditions(const_cast<BoundaryIndexMap&>(probStat[i]->getRhs()->getDOFVector(j)->getBoundaryManager()->getBoundaryConditionMap()));
      }
    }

    // Remove periodic boundaries on elements in mesh.
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh,  -1, Mesh::CALL_EVERY_EL_PREORDER);
    while (elInfo) {
      elInfo->getElement()->deleteElementData(PERIODIC);
      elInfo = stack.traverseNext(elInfo);
    }    

    // Remove periodic vertex associations
    mesh->getPeriodicAssociations().clear();
  }


  void MeshDistributor::removePeriodicBoundaryConditions(BoundaryIndexMap& boundaryMap)
  {
    BoundaryIndexMap::iterator it = boundaryMap.begin();
    while (it != boundaryMap.end()) {
      if (it->second->isPeriodic())
	boundaryMap.erase(it++);
      else
	++it;      
    }    
  }


  void MeshDistributor::checkMeshChange()
  {
    FUNCNAME("MeshDistributor::checkMeshChange()");    

#if (DEBUG != 0)
    debug::writeMesh(feSpace, -1, "before_check_mesh");
#endif

    // === If mesh has not been changed on all ranks, return. ===

    int recvAllValues = 0;
    int sendValue = static_cast<int>(mesh->getChangeIndex() != lastMeshChangeIndex);
    mpiComm.Allreduce(&sendValue, &recvAllValues, 1, MPI_INT, MPI_SUM);

    if (recvAllValues == 0)
      return;

    // === At least one rank mesh has been changed, so the boundaries must be ===
    // === adapted to the new mesh structure.                                 ===

    clock_t first = clock();
    
    do {
      // To check the interior boundaries, the ownership of the boundaries is not 
      // important. Therefore, we add all boundaries to one boundary container.
      RankToBoundMap allBound;

      for (InteriorBoundary::iterator it(myIntBoundary); !it.end(); ++it)
	if ((mesh->getDim() == 2 && it->rankObj.subObj == EDGE) || 
	    (mesh->getDim() == 3 && it->rankObj.subObj == FACE))
 	  allBound[it.getRank()].push_back(*it);

      for (InteriorBoundary::iterator it(otherIntBoundary); !it.end(); ++it)
	if ((mesh->getDim() == 2 && it->rankObj.subObj == EDGE) || 
	    (mesh->getDim() == 3 && it->rankObj.subObj == FACE))
	  allBound[it.getRank()].push_back(*it);

      for (InteriorBoundary::iterator it(periodicBoundary); !it.end(); ++it)
	if ((mesh->getDim() == 2 && it->rankObj.subObj == EDGE) || 
	    (mesh->getDim() == 3 && it->rankObj.subObj == FACE))
 	  allBound[it.getRank()].push_back(*it);	


      // === Check the boundaries and adapt mesh if necessary. ===
#if (DEBUG != 0)
      MSG("Run checkAndAdaptBoundary ...\n");
#endif

      bool meshChanged = checkAndAdaptBoundary(allBound);

      // === Check on all ranks if at least one rank's mesh has changed. ===

      int sendValue = static_cast<int>(!meshChanged);
      recvAllValues = 0;
      mpiComm.Allreduce(&sendValue, &recvAllValues, 1, MPI_INT, MPI_SUM);

#if (DEBUG != 0)
      MSG("Mesh changed on %d ranks!\n", recvAllValues);
#endif
    } while (recvAllValues != 0);

#if (DEBUG != 0)
    debug::writeMesh(feSpace, -1, "mesh");
#endif

    INFO(info, 8)("Parallel mesh adaption needed %.5f seconds\n", 
		  TIME_USED(first, clock()));

    // === Because the mesh has been changed, update the DOF numbering and mappings. ===

    updateLocalGlobalNumbering();


    // === Update periodic mapping, if there are periodic boundaries. ===

    createPeriodicMap();


    // === The mesh has changed, so check if it is required to repartition the mesh. ===

    nTimestepsAfterLastRepartitioning++;

    if (repartitioningAllowed) {
      if (nTimestepsAfterLastRepartitioning >= 20) {
	repartitionMesh();
	nTimestepsAfterLastRepartitioning = 0;
      }
    }
  }

  
  bool MeshDistributor::checkAndAdaptBoundary(RankToBoundMap &allBound)
  {
    FUNCNAME("MeshDistributor::checkAndAdaptBoundary()");

    // === Create mesh structure codes for all ranks boundary elements. ===
       
    std::map<int, MeshCodeVec> sendCodes;
   
    for (RankToBoundMap::iterator it = allBound.begin(); it != allBound.end(); ++it) {

      for (std::vector<AtomicBoundary>::iterator boundIt = it->second.begin();
	   boundIt != it->second.end(); ++boundIt) {
	MeshStructure elCode;
	elCode.init(boundIt->rankObj);
	sendCodes[it->first].push_back(elCode);
      }
    }

    StdMpi<MeshCodeVec> stdMpi(mpiComm, true);
    stdMpi.send(sendCodes);
    stdMpi.recv(allBound);
    stdMpi.startCommunication<uint64_t>(MPI_UNSIGNED_LONG);
 
    // === Compare received mesh structure codes. ===
    
    bool meshFitTogether = true;

    for (RankToBoundMap::iterator it = allBound.begin(); it != allBound.end(); ++it) {
     
      MeshCodeVec &recvCodes = stdMpi.getRecvData()[it->first];
      int i = 0;
      
      for (std::vector<AtomicBoundary>::iterator boundIt = it->second.begin();
	   boundIt != it->second.end(); ++boundIt) {

	MeshStructure elCode;	
	elCode.init(boundIt->rankObj);

	if (elCode.getCode() != recvCodes[i].getCode()) {
	  TEST_EXIT_DBG(refineManager)("Refinement manager is not set correctly!\n");

	  bool b = startFitElementToMeshCode(recvCodes[i], 
					     boundIt->rankObj.el,
					     boundIt->rankObj.subObj,
					     boundIt->rankObj.ithObj, 
					     boundIt->rankObj.elType,
					     boundIt->rankObj.reverseMode);

	  if (b)
	    meshFitTogether = false;	  
 	}

	i++;
      }
    }

    return meshFitTogether;
  }


  bool MeshDistributor::startFitElementToMeshCode(MeshStructure &code, 
						  Element *el, 
						  GeoIndex subObj,
						  int ithObj, 
						  int elType,
						  bool reverseMode)
  {
    FUNCNAME("MeshDistributor::startFitElementToMeshCode()");

    TEST_EXIT_DBG(el)("No element given!\n");

    // If the code is empty, the element does not matter and the function can
    // return without chaning the mesh.
    if (code.empty())
      return false;

    // s0 and s1 are the number of the edge/face in both child of the element,
    // which contain the edge/face the function has to traverse through. If the
    // edge/face is not contained in one of the children, s0 or s1 is -1.
    int s0 = el->getSubObjOfChild(0, subObj, ithObj, elType);
    int s1 = el->getSubObjOfChild(1, subObj, ithObj, elType);

    TEST_EXIT_DBG(s0 != -1 || s1 != -1)("This should not happen!\n");

    bool meshChanged = false;
    Flag traverseFlag = 
      Mesh::CALL_EVERY_EL_PREORDER | Mesh::FILL_NEIGH | Mesh::FILL_BOUND;

    // Test for reverse mode, in which the left and right children of elements
    // are flipped.
    if (reverseMode)
      traverseFlag |= Mesh::CALL_REVERSE_MODE;    


    // === If the edge/face is contained in both children. ===

    if (s0 != -1 && s1 != -1) {
      // Create traverse stack and traverse within the mesh until the element,
      // which should be fitted to the mesh structure code, is reached.
      TraverseStack stack;
      ElInfo *elInfo = stack.traverseFirst(el->getMesh(), -1, traverseFlag);
      while (elInfo && elInfo->getElement() != el)
	elInfo = stack.traverseNext(elInfo);      

      TEST_EXIT_DBG(elInfo->getElement() == el)("This should not happen!\n");

      meshChanged = fitElementToMeshCode(code, stack, subObj, ithObj, reverseMode);
      return meshChanged;
    }


    // === The edge/face is contained in only on of the both children. ===

    if (el->isLeaf()) {

      // If element is leaf and code contains only one leaf element, we are finished.
      if (code.getNumElements() == 1 && code.isLeafElement())
	return false;     

      // Create traverse stack and traverse the mesh to the element.
      TraverseStack stack;
      ElInfo *elInfo = stack.traverseFirst(el->getMesh(), -1, traverseFlag);
      while (elInfo && elInfo->getElement() != el)
	elInfo = stack.traverseNext(elInfo);      

      TEST_EXIT_DBG(elInfo)("This should not happen!\n");

      // Code is not leaf, therefore refine the element.
      el->setMark(1);
      refineManager->setMesh(el->getMesh());
      refineManager->setStack(&stack);
      refineManager->refineFunction(elInfo);
      meshChanged = true;
    }

    Element *child0 = el->getFirstChild();
    Element *child1 = el->getSecondChild();
    if (reverseMode) {
      std::swap(s0, s1);
      std::swap(child0, child1);    
    }

    // === We know that the edge/face is contained in only one of the children. ===
    // === Therefore, traverse the mesh to this children and fit this element   ===
    // === To the mesh structure code.                                          ===

    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(el->getMesh(), -1, traverseFlag);

    if (s0 != -1) {
      while (elInfo && elInfo->getElement() != child0)
	elInfo = stack.traverseNext(elInfo);     

      meshChanged |= fitElementToMeshCode(code, stack, subObj, s0, reverseMode);
    } else {
      while (elInfo && elInfo->getElement() != child1) 
	elInfo = stack.traverseNext(elInfo);      

      meshChanged |= fitElementToMeshCode(code, stack, subObj, s1, reverseMode);
    }


    return meshChanged;
  }


  bool MeshDistributor::fitElementToMeshCode(MeshStructure &code, 
					     TraverseStack &stack,
					     GeoIndex subObj,
					     int ithObj, 
					     bool reverseMode)
  {
    FUNCNAME("MeshDistributor::fitElementToMeshCode()");


    // === Test if there are more elements in stack to check with the code. ===

    ElInfo *elInfo = stack.getElInfo();
    if (!elInfo)
      return false;


    // === Test if code contains a leaf element. If this is the case, the ===
    // === current element is finished. Traverse the mesh to the next     ===
    // === coarser element.                                               ===

    if (code.isLeafElement()) {
      int level = elInfo->getLevel();

      do {
	elInfo = stack.traverseNext(elInfo);
      } while (elInfo && elInfo->getLevel() > level);

      return false;
    }


    bool meshChanged = false;
    Element *el = elInfo->getElement();


    // === If element is leaf (and the code is not), refine the element. ===

    if (el->isLeaf()) {
      TEST_EXIT_DBG(elInfo->getLevel() < 255)("This should not happen!\n");

      el->setMark(1);
      refineManager->setMesh(el->getMesh());
      refineManager->setStack(&stack);
      refineManager->refineFunction(elInfo);
      meshChanged = true;
    }


    // === Continue fitting the mesh structure code to the children of the ===
    // === current element.                                                ===

    int s0 = el->getSubObjOfChild(0, subObj, ithObj, elInfo->getType());
    int s1 = el->getSubObjOfChild(1, subObj, ithObj, elInfo->getType());
    Element *child0 = el->getFirstChild();
    Element *child1 = el->getSecondChild();
    if (reverseMode) {
      std::swap(s0, s1);
      std::swap(child0, child1);
    }

    
    // === Traverse left child. ===

    if (s0 != -1) {
      // The edge/face is contained in the left child, therefore fit this
      // child to the mesh structure code.

      stack.traverseNext(elInfo);
      code.nextElement();
      meshChanged |= fitElementToMeshCode(code, stack, subObj, s0, reverseMode);
      elInfo = stack.getElInfo();
    } else {
      // The edge/face is not contained in the left child. Hence we need
      // to traverse through all subelements of the left child until we get
      // the second child of the current element.

      do {
	elInfo = stack.traverseNext(elInfo);
      } while (elInfo && elInfo->getElement() != child1); 

      TEST_EXIT_DBG(elInfo != NULL)("This should not happen!\n");
    }  

    TEST_EXIT_DBG(elInfo->getElement() == child1)
      ("Got wrong child with idx = %d! Searched for child idx = %d\n",
       elInfo->getElement()->getIndex(), child1->getIndex());


    // === Traverse right child. ===

    if (s1 != -1) {
      // The edge/face is contained in the right child, therefore fit this
      // child to the mesh structure code.

      code.nextElement();
      meshChanged |= fitElementToMeshCode(code, stack, subObj, s1, reverseMode);
    } else {
      // The edge/face is not contained in the right child. Hence we need
      // to traverse through all subelements of the right child until we have
      // finished traversing the current element with all its subelements.

      int level = elInfo->getLevel();

      do {
	elInfo = stack.traverseNext(elInfo);
      } while (elInfo && elInfo->getLevel() > level);
    }


    return meshChanged;
  }

  
  void MeshDistributor::serialize(std::ostream &out, DofContainer &data)
  {    
    int vecSize = data.size();
    SerUtil::serialize(out, vecSize);
    for (int i = 0; i < vecSize; i++) {
      int dofIndex = (*data[i]);
      SerUtil::serialize(out, dofIndex);
    }
  }


  void MeshDistributor::deserialize(std::istream &in, DofContainer &data,
				    std::map<int, const DegreeOfFreedom*> &dofMap)
  {
    FUNCNAME("MeshDistributor::deserialize()");

    int vecSize = 0;
    SerUtil::deserialize(in, vecSize);
    data.clear();
    data.resize(vecSize);
    for (int i = 0; i < vecSize; i++) {
      int dofIndex = 0;
      SerUtil::deserialize(in, dofIndex);

      TEST_EXIT_DBG(dofMap.count(dofIndex) != 0)
	("Dof index could not be deserialized correctly!\n");

      data[i] = dofMap[dofIndex];
    }
  }


  void MeshDistributor::serialize(std::ostream &out, RankToDofContainer &data)
  {
    int mapSize = data.size();
    SerUtil::serialize(out, mapSize);
    for (RankToDofContainer::iterator it = data.begin(); it != data.end(); ++it) {
      int rank = it->first;
      SerUtil::serialize(out, rank);
      serialize(out, it->second);
    }
  }

  
  void MeshDistributor::deserialize(std::istream &in, RankToDofContainer &data,
				    std::map<int, const DegreeOfFreedom*> &dofMap)
  {
    data.clear();

    int mapSize = 0;
    SerUtil::deserialize(in, mapSize);
    for (int i = 0; i < mapSize; i++) {
      int rank = 0;
      SerUtil::deserialize(in, rank);
      deserialize(in, data[rank], dofMap);      
    }
  }


  void MeshDistributor::setInitialElementWeights() 
  {
    FUNCNAME("MeshDistributor::setInitialElementWeights()");

    elemWeights.clear();
      
    std::string filename = "";
    GET_PARAMETER(0, mesh->getName() + "->macro weights", &filename);
    if (filename != "") {
      MSG("Read macro weights from %s\n", filename.c_str());

      std::ifstream infile;
      infile.open(filename.c_str(), std::ifstream::in);
      while (!infile.eof()) {
	int elNum, elWeight;
	infile >> elNum;
	if (infile.eof())
	  break;
	infile >> elWeight;

	elemWeights[elNum] = elWeight;
      }

      infile.close();
    } else {           
      TraverseStack stack;
      ElInfo *elInfo = stack.traverseFirst(mesh, -1, Mesh::CALL_LEAF_EL);
      while (elInfo) {
	elemWeights[elInfo->getElement()->getIndex()] = 1.0;
	elInfo = stack.traverseNext(elInfo);
      }
    }
  }


  void MeshDistributor::repartitionMesh()
  {
    FUNCNAME("MeshDistributor::repartitionMesh()");

    TEST_EXIT(mesh->getNumberOfDOFAdmin() == 1)
      ("Only meshes with one DOFAdmin are supported!\n");

    int nDofs = mesh->getDofAdmin(0).getUsedDofs();
    int repartitioning = 0;

    std::vector<int> nDofsInRank(mpiSize);
    mpiComm.Gather(&nDofs, 1, MPI_INT, &(nDofsInRank[0]), 1, MPI_INT, 0);

    if (mpiRank == 0) {
      int nOverallDofs = 0;
      int minDofs = std::numeric_limits<int>::max();
      int maxDofs = std::numeric_limits<int>::min();
      for (int i = 0; i < mpiSize; i++) {
	nOverallDofs += nDofsInRank[i];
	minDofs = std::min(minDofs, nDofsInRank[i]);
	maxDofs = std::max(maxDofs, nDofsInRank[i]);
      }      
     
      MSG("Overall DOFs: %d    Min DOFs: %d    Max DOFs: %d\n", 
	  nOverallDofs, minDofs, maxDofs);

      if (static_cast<double>(maxDofs) / static_cast<double>(minDofs)) 
	repartitioning = 1;

      mpiComm.Bcast(&repartitioning, 1, MPI_INT, 0);
    } else {
      mpiComm.Bcast(&repartitioning, 1, MPI_INT, 0);
    }


    if (repartitioning == 0)
      return;

    if (repartCounter == 0) {
      std::stringstream oss;
      oss << "partitioning-" << repartCounter << ".vtu";
      DOFVector<double> tmpa(feSpace, "tmp");
      tmpa.set(mpiRank);
      VtkWriter::writeFile(&tmpa, oss.str());
      
      repartCounter++;
    }

    MSG("USED-SIZE A: %d\n", mesh->getDofAdmin(0).getUsedDofs());

    elemWeights.clear();

    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, -1, Mesh::CALL_LEAF_EL);
    while (elInfo) {
      elemWeights[elInfo->getMacroElement()->getIndex()]++;      
      elInfo = stack.traverseNext(elInfo);
    }

    partitioner->useLocalGlobalDofMap(&mapLocalGlobalDofs);
    oldPartitionVec = partitionVec;
    partitioner->partition(&elemWeights, ADAPTIVE_REPART, 1000.0);    

    MacroInfo* minfo = mesh->getMacroFileInfo();
    TEST_EXIT_DBG(minfo)("No macro file info!\n");


    // === Create map that maps macro element indices to pointers to the ===
    // === macro elements.                                               ===

    std::map<int, MacroElement*> elIndexMap;
    for (std::deque<MacroElement*>::iterator it = minfo->mel.begin();
	 it != minfo->mel.end(); ++it)
      elIndexMap[(*it)->getIndex()] = *it;


    // === Create set of all new macro elements this rank will receive from ===
    // === other ranks.                                                     ===

    std::set<MacroElement*> newMacroEl;
    for (std::map<int, std::vector<int> >::iterator it = partitioner->getRecvElements().begin();
	 it != partitioner->getRecvElements().end(); ++it) {
      for (std::vector<int>::iterator elIt = it->second.begin();
	   elIt != it->second.end(); ++elIt) {
	TEST_EXIT_DBG(elIndexMap.count(*elIt) == 1)
	  ("Could not find macro element %d\n", *elIt);
	newMacroEl.insert(elIndexMap[*elIt]);
      }
    }

    std::set<MacroElement*> allMacroEl = newMacroEl;
    for (std::deque<MacroElement*>::iterator it = mesh->firstMacroElement();
	 it != mesh->endOfMacroElements(); ++it)
      allMacroEl.insert(*it);

    // === Add new macro elements to mesh. ===

    for (std::set<MacroElement*>::iterator it = newMacroEl.begin();
	 it != newMacroEl.end(); ++it) {
      MacroElement *mel = *it;
      for (int i = 0; i < mesh->getGeo(NEIGH); i++)
	mel->setNeighbour(i, NULL);

      for (int i = 0; i < mesh->getGeo(VERTEX); i++)
	mel->getElement()->setDof(i, mesh->getDof(VERTEX));

      mesh->getMacroElements().push_back(mel);
    }

    // === Send and receive mesh structure codes. ===

    std::map<int, MeshCodeVec> sendCodes;
    std::map<int, std::vector<std::vector<double> > > sendValues;

    for (std::map<int, std::vector<int> >::iterator it = partitioner->getSendElements().begin();
	 it != partitioner->getSendElements().end(); ++it) {
      for (std::vector<int>::iterator elIt = it->second.begin();
	   elIt != it->second.end(); ++elIt) {
	MeshStructure elCode;
	elCode.init(mesh, *elIt);
	sendCodes[it->first].push_back(elCode);

	for (unsigned int i = 0; i < testVec.size(); i++) {
	  std::vector<double> valVec;
	  elCode.getMeshStructureValues(mesh, *elIt, testVec[i], valVec);
	  sendValues[it->first].push_back(valVec);
	}
      }
    }

    StdMpi<MeshCodeVec> stdMpi(mpiComm, true);
    stdMpi.send(sendCodes);
    stdMpi.recv(partitioner->getRecvElements());
    stdMpi.startCommunication<uint64_t>(MPI_UNSIGNED_LONG);

    StdMpi<std::vector<std::vector<double> > > stdMpi2(mpiComm, true);
    stdMpi2.send(sendValues);
    stdMpi2.recv(partitioner->getRecvElements());
    stdMpi2.startCommunication<double>(MPI_DOUBLE);

    // === Adapte the new macro elements due to the received mesh structure codes. ===

    for (std::map<int, std::vector<int> >::iterator it = partitioner->getRecvElements().begin();
	 it != partitioner->getRecvElements().end(); ++it) {
      MeshCodeVec &recvCodes = stdMpi.getRecvData()[it->first];
      std::vector<std::vector<double> > &recvValues = stdMpi2.getRecvData()[it->first];
    int i = 0;
    int j = 0;

      TEST_EXIT_DBG(recvCodes.size() == it->second.size())
	("Should not happen!\n");

      for (std::vector<int>::iterator elIt = it->second.begin();
	   elIt != it->second.end(); ++elIt) {
	recvCodes[i].fitMeshToStructure(mesh, refineManager, false, *elIt);
	for (unsigned int k = 0; k < testVec.size(); k++)
	  recvCodes[i].setMeshStructureValues(mesh, *elIt, testVec[k], recvValues[j++]);    
	
	i++;
      }
    }


    // === Set correct neighbour information on macro elements. ===

    for (std::set<MacroElement*>::iterator it = allMacroEl.begin();
	 it != allMacroEl.end(); ++it) {
      int elIndex = (*it)->getIndex();
      for (int i = 0; i < mesh->getGeo(NEIGH); i++) {
	int neighIndex = macroElementNeighbours[elIndex][i];
	if (neighIndex == -1 || 
	    partitioner->getElementInRank()[neighIndex] == false)
	  (*it)->setNeighbour(i, NULL);
	else
	  (*it)->setNeighbour(i, elIndexMap[neighIndex]);
      }
    }


    // === Delete macros from mesh. ===

    std::set<MacroElement*> deleteMacroElements;

    for (std::map<int, std::vector<int> >::iterator it = partitioner->getSendElements().begin();
	 it != partitioner->getSendElements().end(); ++it) {
      for (std::vector<int>::iterator elIt = it->second.begin();
	   elIt != it->second.end(); ++elIt) {
	TEST_EXIT_DBG(elIndexMap.count(*elIt) == 1)
	  ("Could not find macro element %d\n", *elIt);

	deleteMacroElements.insert(elIndexMap[*elIt]);
      }
    }

    mesh->removeMacroElements(deleteMacroElements, feSpace);

    if (mpiRank == 0)
      debug::writeElementIndexMesh(mesh, "elementIndexxxx.vtu");


    // === Remove double DOFs. ===
    MeshManipulation meshManipulation(mesh);
    meshManipulation.deleteDoubleDofs(newMacroEl);

    mesh->dofCompress();

    std::stringstream oss;
    oss << "partitioning-" << repartCounter << ".vtu";    
    DOFVector<double> tmpa(feSpace, "tmp");
    tmpa.set(mpiRank);
    VtkWriter::writeFile(&tmpa, oss.str());
    repartCounter++;

    MSG("USED-SIZE B: %d\n", mesh->getDofAdmin(0).getUsedDofs());

#if (DEBUG != 0)
    ParallelDebug::testAllElements(*this);
#endif

    partitioner->fillCoarsePartitionVec(&partitionVec);

    updateInteriorBoundaryInfo();

    updateLocalGlobalNumbering();

   #if (DEBUG != 0)
    MSG("AMDiS runs in debug mode, so make some test ...\n");

    ParallelDebug::testAllElements(*this);
    ParallelDebug::testInteriorBoundary(*this);
    ParallelDebug::testCommonDofs(*this, true);
    ParallelDebug::testGlobalIndexByCoords(*this);

    debug::writeMesh(feSpace, -1, "macro_mesh");   

    MSG("Debug mode tests finished!\n");
#endif
  }


  void MeshDistributor::createInteriorBoundaryInfo()
  {
    FUNCNAME("MeshDistributor::createInteriorBoundaryInfo()");

    createMeshElementData();

    createBoundaryData();

  }


  void MeshDistributor::updateInteriorBoundaryInfo()
  {
    FUNCNAME("MeshDistributor::updateInteriorBoundaryInfo()");

    elObjects.createRankData(partitionVec);

    createBoundaryData();
  }


  void MeshDistributor::createMeshElementData()
  {
    FUNCNAME("MeshDistributor::createMeshElementData()");

   // === Fills macro element data structures. ===

    TraverseStack stack;
    ElInfo *elInfo = 
      stack.traverseFirst(mesh, -1, 
			  Mesh::CALL_LEAF_EL | Mesh::FILL_NEIGH | Mesh::FILL_BOUND);
    while (elInfo) {
      TEST_EXIT_DBG(elInfo->getLevel() == 0)("Should not happen!\n");

      Element *el = elInfo->getElement();
      macroElIndexMap[el->getIndex()] = el;
      macroElIndexTypeMap[el->getIndex()] = elInfo->getType();

      
      // === Add all sub object of the element to the variable elObjects. ===

      for (int i = 0; i < el->getGeo(VERTEX); i++)
	elObjects.addVertex(el->getDof(i, 0), el->getIndex(), i);

      for (int i = 0; i < el->getGeo(EDGE); i++)
	elObjects.addEdge(el->getEdge(i), el->getIndex(), i);      

      for (int i = 0; i < el->getGeo(FACE); i++)
	elObjects.addFace(el->getFace(i), el->getIndex(), i);


      // === Get periodic boundary information. === 

      switch (mesh->getDim()) {
      case 2:
	for (int i = 0; i < el->getGeo(EDGE); i++) {
	  if (mesh->isPeriodicAssociation(elInfo->getBoundary(i))) {
	    DofEdge edge1 = el->getEdge(i);
	    DofEdge edge2 = elInfo->getNeighbour(i)->getEdge(elInfo->getOppVertex(i));	    
	    BoundaryType boundaryType = elInfo->getBoundary(i);

	    periodicEdges[std::make_pair(edge1, edge2)] = boundaryType;


	    for (int j = 0; j < 2; j++) {
	      int ithDofRankObj = el->getVertexOfEdge(i, j);
	      DegreeOfFreedom dof = el->getDof(ithDofRankObj, 0);

	      DegreeOfFreedom neighDof = -1;	  
	      int ithDofNeighObj = -1;	  
	      for (int k = 0; k < 3; k++) {
		if (elInfo->getNeighbour(i)->getDof(k, 0) == 
		    mesh->getPeriodicAssociations(boundaryType)[dof]) {
		  neighDof = elInfo->getNeighbour(i)->getDof(k, 0);
		  ithDofNeighObj = k;
		  break;
		}
	      }
	      TEST_EXIT_DBG(neighDof > -1)("Should not happen!\n");
      		
	      periodicDofs[std::make_pair(dof, neighDof)] = boundaryType;
	      periodicDofAssoc[dof].insert(boundaryType);
	    }
	  }
	}
	break;
      case 3:
	for (int i = 0; i < el->getGeo(FACE); i++) {
	  if (mesh->isPeriodicAssociation(elInfo->getBoundary(i))) {
	    DofFace face1 = el->getFace(i);
	    DofFace face2 = elInfo->getNeighbour(i)->getFace(elInfo->getOppVertex(i));
	    
	    periodicFaces[std::make_pair(face1, face2)] = elInfo->getBoundary(i);
	  }
	}
	break;
      }
	
      elInfo = stack.traverseNext(elInfo);
    }


    // === Create mesh element data for this rank. ===

    elObjects.createRankData(partitionVec);


    // === Search for interectly connected vertices in periodic boundaries. ===

    if (periodicDofs.size() > 0) {
      TEST_EXIT(mesh->getDim() == 2)("Parallel boundaries not yet supported in 3D!\n");

      std::vector<DegreeOfFreedom> twoPeriodicDof;
      for (std::map<DegreeOfFreedom, std::set<BoundaryType> >::iterator it = periodicDofAssoc.begin();
	   it != periodicDofAssoc.end(); ++it) {
	TEST_EXIT_DBG(it->second.size() < 3)("Should not happen!\n");

	if (it->second.size() == 2)
	  twoPeriodicDof.push_back(it->first);
      }

      TEST_EXIT_DBG(twoPeriodicDof.size() == 0 || twoPeriodicDof.size() == 4)
	("Should not happen (%d)!\n", twoPeriodicDof.size());

      if (twoPeriodicDof.size() == 4) {
	for (int i = 0; i < 4; i++) {
	  for (int j = 0; j < 4; j++) {
	    if (i == j) 
	      continue;
	    
	    std::pair<DegreeOfFreedom, DegreeOfFreedom> perDofs = 
	      std::make_pair(twoPeriodicDof[i], twoPeriodicDof[j]);

	    if (periodicDofs.count(perDofs) == 0)
	      periodicDofs[perDofs] = 3;	   	    
	  }
	}
      }
    }
  }


  void MeshDistributor::createBoundaryData()
  {
    FUNCNAME("MeshDistributor::createBoundaryData()");


    // === Clear all relevant data structures, ===

    myIntBoundary.clear();
    otherIntBoundary.clear();
    periodicBoundary.clear();


    // === Create interior boundary data structure. ===

    for (int geoPos = 0; geoPos < mesh->getDim(); geoPos++) {
      GeoIndex geoIndex = INDEX_OF_DIM(geoPos, mesh->getDim());

      while (elObjects.iterate(geoIndex)) {
	std::map<int, ElementObjectData>& objData = elObjects.getIterateData();
	if (objData.count(mpiRank) && objData.size() > 1) {
	  int owner = elObjects.getIterateOwner();
	  ElementObjectData& rankBoundEl = objData[mpiRank];
	  
	  AtomicBoundary bound;	    	    
	  bound.rankObj.el = macroElIndexMap[rankBoundEl.elIndex];
	  bound.rankObj.elIndex = rankBoundEl.elIndex;
	  bound.rankObj.elType = macroElIndexTypeMap[rankBoundEl.elIndex];
	  bound.rankObj.subObj = geoIndex;
	  bound.rankObj.ithObj = rankBoundEl.ithObject;

 	  if (geoIndex == FACE) {
 	    for (int edgeNo = 0; edgeNo < 3; edgeNo++) {
	      int edgeOfFace = 
		bound.rankObj.el->getEdgeOfFace(bound.rankObj.ithObj, edgeNo);

	      bound.rankObj.excludedSubstructures.push_back(std::make_pair(EDGE, edgeOfFace));
	    }
	  }
	  
	  
	  if (owner == mpiRank) {
	    for (std::map<int, ElementObjectData>::iterator it2 = objData.begin();
		 it2 != objData.end(); ++it2) {
	      if (it2->first == mpiRank)
		continue;
	      
	      bound.neighObj.el = macroElIndexMap[it2->second.elIndex];
	      bound.neighObj.elIndex = it2->second.elIndex;
	      bound.neighObj.elType = macroElIndexTypeMap[it2->second.elIndex];
	      bound.neighObj.subObj = geoIndex;
	      bound.neighObj.ithObj = it2->second.ithObject;
	      
	      TEST_EXIT_DBG(rankBoundEl.boundaryType == it2->second.boundaryType)
		("Wrong boundary types: %d %d\n",
		 rankBoundEl.boundaryType,
		 it2->second.boundaryType);
	      
	      bound.type = rankBoundEl.boundaryType;
	      
	      AtomicBoundary& b = myIntBoundary.getNewAtomic(it2->first);
	      b = bound;
	      b.rankObj.setReverseMode(b.neighObj, feSpace);
	    }
	    
	  } else {
	    TEST_EXIT_DBG(objData.count(owner) == 1)
	      ("Should not happen!\n");
	    
	    ElementObjectData& ownerBoundEl = objData[owner];
	    
	    bound.neighObj.el = macroElIndexMap[ownerBoundEl.elIndex];
	    bound.neighObj.elIndex = ownerBoundEl.elIndex;
	    bound.neighObj.elType = -1;
	    bound.neighObj.subObj = geoIndex;
	    bound.neighObj.ithObj = ownerBoundEl.ithObject;
	    
	    TEST_EXIT_DBG(rankBoundEl.boundaryType == ownerBoundEl.boundaryType)
	      ("Should not happen: %d %d\n", 
	       rankBoundEl.boundaryType, ownerBoundEl.boundaryType);
	    
	    bound.type = rankBoundEl.boundaryType;
	    
	    AtomicBoundary& b = otherIntBoundary.getNewAtomic(owner);
	    b = bound;	 
	    b.neighObj.setReverseMode(b.rankObj, feSpace);
	  }
	}
      }
    }
    
    
    // === Create periodic boundary data structure. ===

    for (std::map<std::pair<DegreeOfFreedom, DegreeOfFreedom>, BoundaryType>::iterator it = periodicDofs.begin();
	 it != periodicDofs.end(); ++it) {
      if (elObjects.isInRank(it->first.first, mpiRank) == false)
	continue;

      for (std::map<int, ElementObjectData>::iterator elIt = elObjects.getElementsInRank(it->first.second).begin();
	   elIt != elObjects.getElementsInRank(it->first.second).end(); ++elIt) {

	int otherElementRank = elIt->first;
	ElementObjectData& perDofEl0 = elObjects.getElementsInRank(it->first.first)[mpiRank];
	ElementObjectData& perDofEl1 = elIt->second;

	AtomicBoundary bound;
	bound.rankObj.el = macroElIndexMap[perDofEl0.elIndex];
	bound.rankObj.elIndex = perDofEl0.elIndex;
	bound.rankObj.elType = macroElIndexTypeMap[perDofEl0.elIndex];
	bound.rankObj.subObj = VERTEX;
	bound.rankObj.ithObj = perDofEl0.ithObject;

	bound.neighObj.el = macroElIndexMap[perDofEl1.elIndex];
	bound.neighObj.elIndex = perDofEl1.elIndex;
	bound.neighObj.elType = -1;
	bound.neighObj.subObj = VERTEX;
	bound.neighObj.ithObj = perDofEl1.ithObject;

	bound.type = it->second;

	AtomicBoundary& b = periodicBoundary.getNewAtomic(otherElementRank);
	b = bound;
      }
    }



    for (std::map<std::pair<DofEdge, DofEdge>, BoundaryType>::iterator it = periodicEdges.begin();
	 it != periodicEdges.end(); ++it) {
      int perEdgeOwner0 = elObjects.getOwner(it->first.first);
      int perEdgeOwner1 = elObjects.getOwner(it->first.second);

      if (perEdgeOwner0 != mpiRank)
	continue;

      TEST_EXIT_DBG(perEdgeOwner0 != perEdgeOwner1)("Should not happen!\n");
      TEST_EXIT_DBG(elObjects.getElements(it->first.first).size() == 1)
	("Should not happen!\n");
      TEST_EXIT_DBG(elObjects.getElements(it->first.second).size() == 1)
	("Should not happen!\n");
      
      int otherElementRank = perEdgeOwner1;
      ElementObjectData& perEdgeEl0 = elObjects.getElements(it->first.first)[0];
      ElementObjectData& perEdgeEl1 = elObjects.getElements(it->first.second)[0];
      
      AtomicBoundary bound;	    	    
      bound.rankObj.el = macroElIndexMap[perEdgeEl0.elIndex];
      bound.rankObj.elIndex = perEdgeEl0.elIndex;
      bound.rankObj.elType = macroElIndexTypeMap[perEdgeEl0.elIndex];
      bound.rankObj.subObj = EDGE;
      bound.rankObj.ithObj = perEdgeEl0.ithObject;
      
      bound.neighObj.el = macroElIndexMap[perEdgeEl1.elIndex];
      bound.neighObj.elIndex = perEdgeEl1.elIndex;
      bound.neighObj.elType = -1;
      bound.neighObj.subObj = EDGE;
      bound.neighObj.ithObj = perEdgeEl1.ithObject;
      
      bound.type = it->second;
      
      AtomicBoundary& b = periodicBoundary.getNewAtomic(otherElementRank);
      b = bound;
      
      if (mpiRank > otherElementRank)
	b.rankObj.setReverseMode(b.neighObj, feSpace);
      else
	b.neighObj.setReverseMode(b.rankObj, feSpace);     
    }



    // === Once we have this information, we must care about the order of the atomic ===
    // === bounds in the three boundary handling object. Eventually all the bound-   ===
    // === aries have to be in the same order on both ranks that share the bounday.  ===

    StdMpi<std::vector<AtomicBoundary> > stdMpi(mpiComm);
    stdMpi.send(myIntBoundary.boundary);
    stdMpi.recv(otherIntBoundary.boundary);
    stdMpi.startCommunication<int>(MPI_INT);

    // === The information about all neighbouring boundaries has been received. So ===
    // === the rank tests if its own atomic boundaries are in the same order. If   ===
    // === not, the atomic boundaries are swaped to the correct order.             ===

    for (RankToBoundMap::iterator rankIt = otherIntBoundary.boundary.begin();
	 rankIt != otherIntBoundary.boundary.end(); ++rankIt) {

      // === We have received from rank "rankIt->first" the ordered list of element ===
      // === indices. Now, we have to sort the corresponding list in this rank to   ===
      // === get the same order.                                                    ===
     
      for (unsigned int j = 0; j < rankIt->second.size(); j++) {

	// If the expected object is not at place, search for it.

	BoundaryObject &recvedBound = stdMpi.getRecvData()[rankIt->first][j].rankObj;

	if ((rankIt->second)[j].neighObj != recvedBound) {
	  unsigned int k = j + 1;

	  for (; k < rankIt->second.size(); k++)
 	    if ((rankIt->second)[k].neighObj == recvedBound)
	      break;

	  // The element must always be found, because the list is just in another order.
	  TEST_EXIT_DBG(k < rankIt->second.size())("Should never happen!\n");

	  // Swap the current with the found element.
	  AtomicBoundary tmpBound = (rankIt->second)[k];
	  (rankIt->second)[k] = (rankIt->second)[j];
	  (rankIt->second)[j] = tmpBound;	
	}
      }
    }


    // === Do the same for the periodic boundaries. ===

    if (periodicBoundary.boundary.size() > 0) {
      stdMpi.clear();

      InteriorBoundary sendBounds, recvBounds;     
      for (RankToBoundMap::iterator rankIt = periodicBoundary.boundary.begin();
	   rankIt != periodicBoundary.boundary.end(); ++rankIt) {

	TEST_EXIT_DBG(rankIt->first != mpiRank)
	  ("It is not allowed to have an interior boundary within a rank partition!\n");

	if (rankIt->first < mpiRank)
	  sendBounds.boundary[rankIt->first] = rankIt->second;
	else
	  recvBounds.boundary[rankIt->first] = rankIt->second;
      }

      stdMpi.send(sendBounds.boundary);
      stdMpi.recv(recvBounds.boundary);
      stdMpi.startCommunication<int>(MPI_INT);

      for (RankToBoundMap::iterator rankIt = periodicBoundary.boundary.begin();
	   rankIt != periodicBoundary.boundary.end(); ++rankIt) {
	if (rankIt->first <= mpiRank)
	  continue;
	  
	for (unsigned int j = 0; j < rankIt->second.size(); j++) {
	  
	  BoundaryObject &recvedBound = stdMpi.getRecvData()[rankIt->first][j].rankObj;
	  
	  if (periodicBoundary.boundary[rankIt->first][j].neighObj != recvedBound) {    
	    unsigned int k = j + 1;	    
	    for (; k < rankIt->second.size(); k++)
	      if (periodicBoundary.boundary[rankIt->first][k].neighObj == recvedBound)
		break;
	    
	    // The element must always be found, because the list is just in 
	    // another order.
	    TEST_EXIT_DBG(k < rankIt->second.size())("Should never happen!\n");
	    
	    // Swap the current with the found element.
	    AtomicBoundary tmpBound = (rankIt->second)[k];
	    (rankIt->second)[k] = (rankIt->second)[j];
	    (rankIt->second)[j] = tmpBound;	
	  }  	  
	}
      }     
    } // periodicBoundary.boundary.size() > 0
  }


  void MeshDistributor::removeMacroElements()
  {
    FUNCNAME("MeshDistributor::removeMacroElements()");

    std::set<MacroElement*> macrosToRemove;
    for (std::deque<MacroElement*>::iterator it = mesh->firstMacroElement();
	 it != mesh->endOfMacroElements(); ++it)
      if (partitioner->getElementInRank()[(*it)->getIndex()] == false)
	macrosToRemove.insert(*it);    

    mesh->removeMacroElements(macrosToRemove, feSpace);
  }


  void MeshDistributor::updateLocalGlobalNumbering()
  {
    FUNCNAME("MeshDistributor::updateLocalGlobalNumbering()");

    mesh->dofCompress();

#if (DEBUG != 0)
    debug::ElementIdxToDofs elMap;
    debug::createSortedDofs(mesh, elMap);   
#endif

    typedef std::set<const DegreeOfFreedom*> DofSet;

    // === Get all DOFs in ranks partition. ===

    ElementDofIterator elDofIt(feSpace);
    DofSet rankDofSet;
    
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, -1, Mesh::CALL_LEAF_EL);
    while (elInfo) {
      elDofIt.reset(elInfo->getElement());
      do {
	rankDofSet.insert(elDofIt.getDofPtr());
      } while(elDofIt.next());

      elInfo = stack.traverseNext(elInfo);
    }


    DofContainer rankDofs;
    for (DofSet::iterator it = rankDofSet.begin(); it != rankDofSet.end(); ++it)
      rankDofs.push_back(*it);    
    sort(rankDofs.begin(), rankDofs.end(), cmpDofsByValue);
    int nRankAllDofs = rankDofs.size();


    // === Traverse on interior boundaries and move all not ranked owned DOFs from ===
    // === rankDofs to boundaryDofs.                                               ===

    RankToDofContainer oldSendDofs = sendDofs;
    RankToDofContainer oldRecvDofs = recvDofs;

    sendDofs.clear();
    recvDofs.clear();


    for (InteriorBoundary::iterator it(myIntBoundary); !it.end(); ++it) {
      DofContainer dofs;	
      it->rankObj.el->getVertexDofs(feSpace, it->rankObj, dofs);
      it->rankObj.el->getNonVertexDofs(feSpace, it->rankObj, dofs);

      for (unsigned int i = 0; i < dofs.size(); i++)
	sendDofs[it.getRank()].push_back(dofs[i]);
    }


    for (InteriorBoundary::iterator it(otherIntBoundary); !it.end(); ++it) {
      DofContainer dofs;
      it->rankObj.el->getVertexDofs(feSpace, it->rankObj, dofs);
      it->rankObj.el->getNonVertexDofs(feSpace, it->rankObj, dofs);

      for (unsigned int i = 0; i < dofs.size(); i++) {
	DofContainer::iterator eraseIt = 
	  find(rankDofs.begin(), rankDofs.end(), dofs[i]);
	if (eraseIt != rankDofs.end()) 
	  rankDofs.erase(eraseIt);

	recvDofs[it.getRank()].push_back(dofs[i]);
      }
    }


    nRankDofs = rankDofs.size();

    // === Get starting position for global rank dof ordering. ====

    mpiComm.Scan(&nRankDofs, &rstart, 1, MPI_INT, MPI_SUM);
    rstart -= nRankDofs;


    // === Calculate number of overall DOFs of all partitions. ===

    nOverallDofs = 0;
    mpiComm.Allreduce(&nRankDofs, &nOverallDofs, 1, MPI_INT, MPI_SUM);

    // First, we set all dofs in ranks partition to be owend by the rank. Later,
    // the dofs in ranks partition that are owned by other rank are set to false.
    isRankDof.clear();
    for (int i = 0; i < nRankAllDofs; i++)
      isRankDof[i] = true;    

    // Stores for all rank owned dofs a new global index.
    DofIndexMap rankDofsNewGlobalIndex;
    for (int i = 0; i < nRankDofs; i++)
      rankDofsNewGlobalIndex[rankDofs[i]] = i + rstart;


    // === Send new DOF indices. ===

#if (DEBUG != 0)
    ParallelDebug::testDofContainerCommunication(*this, sendDofs, recvDofs);
#endif
    
    int i = 0;
    StdMpi<std::vector<DegreeOfFreedom> > stdMpi(mpiComm, false);
    for (RankToDofContainer::iterator sendIt = sendDofs.begin();
	 sendIt != sendDofs.end(); ++sendIt, i++) {
      stdMpi.getSendData(sendIt->first).resize(0);
      stdMpi.getSendData(sendIt->first).reserve(sendIt->second.size());
      for (DofContainer::iterator dofIt = sendIt->second.begin();
	   dofIt != sendIt->second.end(); ++dofIt)
	stdMpi.getSendData(sendIt->first).push_back(rankDofsNewGlobalIndex[*dofIt]);
    }
    stdMpi.updateSendDataSize();
    stdMpi.recv(recvDofs);
    stdMpi.startCommunication<int>(MPI_INT);


    for (RankToDofContainer::iterator recvIt = recvDofs.begin();
	 recvIt != recvDofs.end(); ++recvIt) {      
      int j = 0;
      for (DofContainer::iterator dofIt = recvIt->second.begin();
	   dofIt != recvIt->second.end(); ++dofIt) {
	rankDofsNewGlobalIndex[*dofIt] = stdMpi.getRecvData(recvIt->first)[j++];
	isRankDof[**dofIt] = false;
      }
    }


    // === Create now the local to global index and local to dof index mappings.  ===

    mapLocalGlobalDofs.clear();
    mapLocalDofIndex.clear();

    for (DofIndexMap::iterator dofIt = rankDofsNewGlobalIndex.begin();
	 dofIt != rankDofsNewGlobalIndex.end(); ++dofIt)
      mapLocalGlobalDofs[*(dofIt->first)] = dofIt->second;   

    for (int i = 0; i < nRankDofs; i++)
      mapLocalDofIndex[i] = *(rankDofs[i]);    


    // === Update dof admins due to new number of dofs. ===
  
    lastMeshChangeIndex = mesh->getChangeIndex();
    
#if (DEBUG != 0)
    std::stringstream oss;
    oss << "elementIndex-" << mpiRank << ".vtu";

    debug::writeElementIndexMesh(mesh, oss.str());
    debug::testSortedDofs(mesh, elMap);
    ParallelDebug::testCommonDofs(*this, true);
    ParallelDebug::testGlobalIndexByCoords(*this);
    ParallelDebug::writeDebugFile(*this, "mpi-dbg", "dat");
    
#if 0
    MSG("------------- Debug information -------------\n");
    MSG("nRankDofs = %d\n", nRankDofs);
    MSG("nOverallDofs = %d\n", nOverallDofs);
    MSG("rstart %d\n", rstart);
    /*
    for (RankToDofContainer::iterator it = sendDofs.begin(); 
	 it != sendDofs.end(); ++it) {
      MSG("send %d DOFs to rank %d\n", it->second.size(), it->first);

      for (unsigned int i = 0; i < it->second.size(); i++)
	MSG("%d send DOF: %d\n", i, *(it->second[i]));
    }
    for (RankToDofContainer::iterator it = recvDofs.begin();
	 it != recvDofs.end(); ++it) {
      MSG("recv %d DOFs from rank %d\n", it->second.size(), it->first);

      for (unsigned int i = 0; i < it->second.size(); i++)
	MSG("%d recv DOF: %d\n", i, *(it->second[i]));
    }
    */
#endif
#endif
  }


  void MeshDistributor::createPeriodicMap()
  {
    FUNCNAME("MeshDistributor::createPeriodicMap()");

    if (periodicBoundary.boundary.size() == 0)
      return;

    // Clear all periodic dof mappings calculated before. We do it from scratch.
    periodicDof.clear();
    periodicDofAssociations.clear();

    StdMpi<std::vector<int> > stdMpi(mpiComm, false);

    // === Each rank traverse its periodic boundaries and sends the dof indices ===
    // === to the rank "on the other side" of the periodic boundary.            ===

    RankToDofContainer rankPeriodicDofs;
    std::map<int, std::vector<int> > rankToDofType;

    for (RankToBoundMap::iterator it = periodicBoundary.boundary.begin();
	 it != periodicBoundary.boundary.end(); ++it) {
      TEST_EXIT_DBG(it->first != mpiRank)
	("Periodic interior boundary within the rank itself is not possible!\n");

      // Create dof indices on the boundary. 
      DofContainer& dofs = rankPeriodicDofs[it->first];
      for (std::vector<AtomicBoundary>::iterator boundIt = it->second.begin();
	   boundIt != it->second.end(); ++boundIt) {
	int nDofs = dofs.size();

	boundIt->rankObj.el->getVertexDofs(feSpace, boundIt->rankObj, dofs);
	boundIt->rankObj.el->getNonVertexDofs(feSpace, boundIt->rankObj, dofs);

	for (unsigned int i = 0; i < (dofs.size() - nDofs); i++)
	  rankToDofType[it->first].push_back(boundIt->type);
      }

      // Send the global indices to the rank on the other side.
      stdMpi.getSendData(it->first).reserve(dofs.size());
      for (unsigned int i = 0; i < dofs.size(); i++)
	stdMpi.getSendData(it->first).push_back(mapLocalGlobalDofs[*(dofs[i])]);      

      // Receive from this rank the same number of dofs.
      stdMpi.recv(it->first, dofs.size());
    }

    stdMpi.updateSendDataSize();
    stdMpi.startCommunication<int>(MPI_INT);


    // === The rank has received the dofs from the rank on the other side of ===
    // === the boundary. Now it can use them to create the mapping between   ===
    // === the periodic dofs in this rank and the corresponding periodic     ===
    // === dofs from the other ranks.                                        ===


    std::map<DegreeOfFreedom, std::set<int> > dofFromRank;

    for (RankToBoundMap::iterator it = periodicBoundary.boundary.begin();
	 it != periodicBoundary.boundary.end(); ++it) {
      DofContainer& dofs = rankPeriodicDofs[it->first];
      std::vector<int>& types = rankToDofType[it->first];

      TEST_EXIT_DBG(dofs.size() == types.size())("Should not happen!\n");

      // Added the received dofs to the mapping.
      for (unsigned int i = 0; i < dofs.size(); i++) {
	int globalDofIndex = mapLocalGlobalDofs[*(dofs[i])];
	int mapGlobalDofIndex = stdMpi.getRecvData(it->first)[i];
	BoundaryType type = types[i];

	// Check if this global dof with the corresponding boundary type was
	// not added before by another periodic boundary from other rank.
	if (periodicDofAssociations[globalDofIndex].count(type) == 0) {
	  periodicDof[type][globalDofIndex] = mapGlobalDofIndex;
	  periodicDofAssociations[globalDofIndex].insert(type);
	  dofFromRank[globalDofIndex].insert(it->first);
	}
      }
    }


#if (DEBUG != 0)
    for (std::map<int, std::set<BoundaryType> >::iterator it = periodicDofAssociations.begin();
	 it != periodicDofAssociations.end(); ++it) {
      int nAssoc = it->second.size();
      TEST_EXIT_DBG(nAssoc == 1 || nAssoc == 3)
 	("Should not happen! DOF %d has %d periodic associations!\n", 
 	 it->first, nAssoc);
    }
#endif
  }


  void MeshDistributor::createMacroElementInfo()
  {
    FUNCNAME("MeshDistributor::createMacroElementInfo()");

    for (std::deque<MacroElement*>::iterator it = mesh->firstMacroElement();
	 it != mesh->endOfMacroElements(); ++it) {
      macroElementNeighbours[(*it)->getIndex()].resize(mesh->getGeo(NEIGH));
      for (int i = 0; i < mesh->getGeo(NEIGH); i++)
	macroElementNeighbours[(*it)->getIndex()][i] = 
	  (*it)->getNeighbour(i) ? (*it)->getNeighbour(i)->getIndex() : -1;
    }     
  }


  void MeshDistributor::serialize(std::ostream &out)
  {
    SerUtil::serialize(out, elemWeights);
    SerUtil::serialize(out, partitionVec);
    SerUtil::serialize(out, oldPartitionVec);
    SerUtil::serialize(out, nRankDofs);
    SerUtil::serialize(out, nOverallDofs);

    elObjects.serialize(out);

    SerUtil::serialize(out, periodicDofs);
    SerUtil::serialize(out, periodicEdges);
    SerUtil::serialize(out, periodicFaces);

    myIntBoundary.serialize(out);
    otherIntBoundary.serialize(out);
    periodicBoundary.serialize(out);

    serialize(out, sendDofs);
    serialize(out, recvDofs);

    SerUtil::serialize(out, mapLocalGlobalDofs);
    SerUtil::serialize(out, mapLocalDofIndex);
    SerUtil::serialize(out, isRankDof);

    serialize(out, periodicDof);
    serialize(out, periodicDofAssociations);

    SerUtil::serialize(out, rstart);
    SerUtil::serialize(out, macroElementNeighbours);
  }


  void MeshDistributor::deserialize(std::istream &in)
  {
    FUNCNAME("MeshDistributor::deserialize()");

    SerUtil::deserialize(in, elemWeights);
    SerUtil::deserialize(in, partitionVec);
    SerUtil::deserialize(in, oldPartitionVec);
    SerUtil::deserialize(in, nRankDofs);    
    SerUtil::deserialize(in, nOverallDofs);

    // Create two maps: one from from element indices to the corresponding element 
    // pointers, and one map from Dof indices to the corresponding dof pointers.
    std::map<int, Element*> elIndexMap;
    std::map<int, const DegreeOfFreedom*> dofMap;
    ElementDofIterator elDofIter(feSpace);
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, -1, Mesh::CALL_EVERY_EL_PREORDER);
    while (elInfo) {
      Element *el = elInfo->getElement();
      elIndexMap[el->getIndex()] = el;

      if (el->isLeaf()) {
	elDofIter.reset(el);
	do {
	  dofMap[elDofIter.getDof()] = elDofIter.getDofPtr();
	} while (elDofIter.next());      
      }

      elInfo = stack.traverseNext(elInfo);
    }

    elObjects.deserialize(in);
   
    SerUtil::deserialize(in, periodicDofs);
    SerUtil::deserialize(in, periodicEdges);
    SerUtil::deserialize(in, periodicFaces);
   
    myIntBoundary.deserialize(in, elIndexMap);
    otherIntBoundary.deserialize(in, elIndexMap);
    periodicBoundary.deserialize(in, elIndexMap);

    deserialize(in, sendDofs, dofMap);
    deserialize(in, recvDofs, dofMap);

    SerUtil::deserialize(in, mapLocalGlobalDofs);
    SerUtil::deserialize(in, mapLocalDofIndex);
    SerUtil::deserialize(in, isRankDof);

    deserialize(in, periodicDof);
    deserialize(in, periodicDofAssociations);

    SerUtil::deserialize(in, rstart);
    SerUtil::deserialize(in, macroElementNeighbours);

    deserialized = true;
  }


  void MeshDistributor::serialize(std::ostream &out, PeriodicDofMap &data)
  {
    int mapSize = data.size();
    SerUtil::serialize(out, mapSize);

    for (PeriodicDofMap::iterator it = data.begin(); it != data.end(); ++it) {
      int type = it->first;
      DofMapping dofMap = it->second;

      SerUtil::serialize(out, type);
      SerUtil::serialize(out, dofMap);
    }
  }


  void MeshDistributor::serialize(std::ostream &out, std::map<int, std::set<int> >& data)
  {
    int mapSize = data.size();
    SerUtil::serialize(out, mapSize);

    for (std::map<int, std::set<int> >::iterator it = data.begin(); it != data.end(); ++it) {
      int dof = it->first;
      std::set<int> typeSet = it->second;

      SerUtil::serialize(out, dof);
      SerUtil::serialize(out, typeSet);
    }
  }


  void MeshDistributor::deserialize(std::istream &in, PeriodicDofMap &data)
  {
    data.clear();

    int mapSize = 0;
    SerUtil::deserialize(in, mapSize);

    for (int i = 0; i < mapSize; i++) {
      int type;
      DofMapping dofMap;

      SerUtil::deserialize(in, type);
      SerUtil::deserialize(in, dofMap);

      data[type] = dofMap;
    }
  }


  void MeshDistributor::deserialize(std::istream &in, 
				    std::map<int, std::set<int> >& data)
  {
    data.clear();

    int mapSize = 0;
    SerUtil::deserialize(in, mapSize);

    for (int i = 0; i < mapSize; i++) {
      int dof;
      std::set<int> typeSet;

      SerUtil::deserialize(in, dof);
      SerUtil::deserialize(in, typeSet);

      data[dof] = typeSet;
    }
  }

 
  void MeshDistributor::writePartitioningMesh(std::string filename)
  {
    FUNCNAME("MeshDistributor::writePartitioningMesh()");

    std::map<int, double> vec;    
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, -1, 
					 Mesh::CALL_LEAF_EL | Mesh::FILL_COORDS);
    
    while (elInfo) {		  
      int index = elInfo->getElement()->getIndex();
      vec[index] = partitionVec[index];
      elInfo = stack.traverseNext(elInfo);
    }

    ElementFileWriter::writeFile(vec, mesh, filename);
  }

}
